"""Constants for the Wolf SmartSet Service integration."""

DOMAIN = "wolflink"

COORDINATOR = "coordinator"
PARAMETERS = "parameters"
DEVICE_ID = "device_id"
DEVICE_GATEWAY = "device_gateway"
DEVICE_NAME = "device_name"

STATES = {
    "Ein": "ein",
    "Deaktiviert": "deaktiviert",
    "Aus": "aus",
    "Standby": "standby",
    "Auto": "auto",
    "Permanent": "permanent",
    "Initialisierung": "initialisierung",
    "Antilegionellenfunktion": "antilegionellenfunktion",
    "Fernschalter ein": "fernschalter_ein",
    "1x Warmwasser": "1_x_warmwasser",
    "Bereit, keine Ladung": "bereit_keine_ladung",
    "Solarbetrieb": "solarbetrieb",
    "Reduzierter Betrieb": "reduzierter_betrieb",
    "SmartHome": "smart_home",
    "SmartGrid": "smart_grid",
    "Ruhekontakt": "ruhekontakt",
    "Vorspülen": "vorspulen",
    "Zünden": "zunden",
    "Stabilisierung": "stabilisierung",
    "Ventilprüfung": "ventilprufung",
    "Nachspülen": "nachspulen",
    "Softstart": "softstart",
    "Taktsperre": "taktsperre",
    "Betrieb ohne Brenner": "betrieb_ohne_brenner",
    "Abgasklappe": "abgasklappe",
    "Störung": "storung",
    "Gradienten Überwachung": "gradienten_uberwachung",
    "Gasdruck": "gasdruck",
    "Spreizung hoch": "spreizung_hoch",
    "Spreizung KF": "spreizung_kf",
    "Test": "test",
    "Start": "start",
    "Frost Heizkreis": "frost_heizkreis",
    "Frost Warmwasser": "frost_warmwasser",
    "Schornsteinfeger": "schornsteinfeger",
    "Kombibetrieb": "kombibetrieb",
    "Parallelbetrieb": "parallelbetrieb",
    "Warmwasserbetrieb": "warmwasserbetrieb",
    "Warmwassernachlauf": "warmwassernachlauf",
    "Mindest-Kombizeit": "mindest_kombizeit",
    "Heizbetrieb": "heizbetrieb",
    "Nachlauf Heizkreispumpe": "nachlauf_heizkreispumpe",
    "Frostschutz": "frostschutz",
    "Kaskadenbetrieb": "kaskadenbetrieb",
    "GLT-Betrieb": "glt_betrieb",
    "Kalibration": "kalibration",
    "Kalibration Heizbetrieb": "kalibration_heizbetrieb",
    "Kalibration Warmwasserbetrieb": "kalibration_warmwasserbetrieb",
    "Kalibration Kombibetrieb": "kalibration_kombibetrieb",
    "Warmwasser Schnellstart": "warmwasser_schnellstart",
    "Externe Deaktivierung": "externe_deaktivierung",
    "Heizung": "heizung",
    "Warmwasser": "warmwasser",
    "Kombigerät": "kombigerat",
    "Kombigerät mit Solareinbindung": "kombigerat_mit_solareinbindung",
    "Heizgerät mit Speicher": "heizgerat_mit_speicher",
    "Nur Heizgerät": "nur_heizgerat",
    "Aktiviert": "ktiviert",
    "Sparen": "sparen",
    "Estrichtrocknung": "estrichtrocknung",
    "Telefonfernschalter": "telefonfernschalter",
    "Partymodus": "partymodus",
    "Urlaubsmodus": "urlaubsmodus",
    "Automatik ein": "automatik_ein",
    "Automatik aus": "automatik_aus",
    "Permanentbetrieb": "permanentbetrieb",
    "Sparbetrieb": "sparbetrieb",
    "AutoOnCool": "auto_on_cool",
    "AutoOffCool": "auto_off_cool",
    "PermCooling": "perm_cooling",
    "Absenkbetrieb": "absenkbetrieb",
    "Eco": "eco",
    "Absenkstop": "absenkstop",
    "AT Abschaltung": "at_abschaltung",
    "RT Abschaltung": "rt_abschaltung",
    "AT Frostschutz": "at_frostschutz",
    "RT Frostschutz": "rt_frostschutz",
    "DHWPrior": "dhw_prior",
    "Cooling": "cooling",
    "TPW": "tpw",
    "Warmwasservorrang": "warmwasservorrang",
}
