"""Const file for the BMW Connected Drive integration."""
ATTRIBUTION = "Data provided by BMW Connected Drive"

CONF_REGION = "region"
CONF_ALLOWED_REGIONS = ["china", "north_america", "rest_of_world"]
CONF_READ_ONLY = "read_only"
CONF_USE_LOCATION = "use_location"

CONF_ACCOUNT = "account"

DATA_HASS_CONFIG = "hass_config"
DATA_ENTRIES = "entries"
