"""Constants for the Ondilo ICO integration."""

DOMAIN = "ondilo_ico"

OAUTH2_AUTHORIZE = "https://interop.ondilo.com/oauth2/authorize"
OAUTH2_TOKEN = "https://interop.ondilo.com/oauth2/token"
OAUTH2_CLIENTID = "customer_api"
OAUTH2_CLIENTSECRET = ""
